-- Public Domain software.
-- Euphoria 2.5 -- unlimited alphabet bilingual interpreter.
-- Any code page for the names.
-- English and Russian keywords, builtins, standard routines.
-- English  or Russian interpreter messages.
-- A Euphoria Interpreter written 100% in Euphoria,
-- reconstruction by Igor N. Kachan, 2005 May.
-- Report the problems to  kinz@peterlink.ru


-- Compile-time Error Handling

include machine.e

-- 22 + the final "too many" warning
constant MAX_WARNINGS = 23
global constant STDIN = 0, STDERR = 2

integer Errors 
Errors = 0   -- number of errors detected during compile 

global integer TempErrFile
global sequence TempErrName
global object ThisLine        -- current line of source (or -1)
global integer bp             -- input line index of next character 

global sequence warning_list
warning_list = {}

global procedure screen_output(integer f, object msg) -- kinz
-- output messages to the screen or a file
    if atom(msg) then 
       puts(f, messages[msg][LANG])
    else
       puts(f, msg)
    end if   
end procedure

global procedure Warning(object msg) -- kinz
-- add a warning message to the list
    sequence p
    if OpWarning and length(warning_list) < MAX_WARNINGS then
	if length(warning_list) = MAX_WARNINGS-1 then
	    p = messages[016][LANG]
	    --"Too many warnings - no more will be issued"
	else 
	    if atom(msg) then
	      p = sprintf(messages[017][LANG], {messages[msg][LANG]})
			  --"Warning: %s\n"
	    else
	      p = sprintf(messages[017][LANG], {msg})
	    end if
	    if find(p, warning_list) then
		return -- duplicate
	    end if
	end if
	warning_list = append(warning_list, p)          
    end if
end procedure

global function ShowWarnings(integer errfile)
-- print the warnings to the screen (or ex.err)
    for i = 1 to length(warning_list) do
	if errfile = 0 then
	    screen_output(STDERR, warning_list[i])
	else
	    puts(errfile, warning_list[i])
	end if
    end for
    return length(warning_list)
end function

global procedure Cleanup(integer status)
-- clean things up before quitting
    integer w
    
    w = ShowWarnings(0) 
    
    if not TRANSLATE and 
       (BIND or EWINDOWS or ELINUX) and 
       (w or Errors) then
	screen_output(STDERR, 018)--"\nPress Enter\n"
	if getc(0) then -- prompt
	end if
    end if
    
    abort(status)
end procedure

global procedure OpenErrFile() -- kinz
-- open the error diagnostics file - normally "ex.err"
    TempErrFile = open(TempErrName, "w")
    if TempErrFile = -1 then
	if length(TempErrName) > 0 then
	    screen_output(STDERR, 019)--"Can't create error message file: "
	    screen_output(STDERR, TempErrName)
	    screen_output(STDERR, "\n")
	end if
	abort(1) --Cleanup(1)
    end if
end procedure

procedure ShowErr(integer f) -- kinz
-- Show place where syntax error occurred  
    if length(file_name) = 0 then
	return
    end if
    if ThisLine[1] = END_OF_FILE_CHAR then
	screen_output(f, 020)--"<end-of-file>\n"
    else
	screen_output(f, ThisLine)
    end if
    
    for i = 1 to bp-2 do -- bp-1 points to last character read
	if ThisLine[i] = '\t' then
	    screen_output(f, "\t")
	else
	    screen_output(f, " ")
	end if
    end for
    screen_output(f, "^\n\n") 
end procedure

global procedure CompileErr(object msg) -- kinz
-- Handle fatal compilation errors 
    sequence errmsg
    
    Errors += 1
    if length(file_name) then
	if atom(msg) then
	  errmsg = sprintf("%.99s:%d\n%s\n", {file_name[current_file_no], 
					      line_number, messages[msg][LANG]})
	else             
	  errmsg = sprintf("%.99s:%d\n%s\n", {file_name[current_file_no], 
					      line_number, msg})
	end if
    else
	if atom(msg) then
	  errmsg = messages[msg][LANG]
	else
	  errmsg = msg
	end if
    end if
	-- try to open err file *before* displaying diagnostics on screen
    OpenErrFile() -- exits if error filename is ""
    screen_output(STDERR, errmsg)
    ShowErr(STDERR)
    
    puts(TempErrFile, errmsg) 
    
    ShowErr(TempErrFile) 
    
    if ShowWarnings(TempErrFile) then
    end if
    
    puts(TempErrFile, "\n")
    
    close(TempErrFile)
    Cleanup(1)
end procedure

procedure not_supported_compile(object feature) -- kinz
-- report feature not supported
 if atom(feature) then
    CompileErr(sprintf(messages[021][LANG],
		     --"%s is not supported in Euphoria for %s"
	       {messages[feature][LANG], version_name[LANG]}))
 else
    CompileErr(sprintf(messages[021][LANG], 
		     --"%s is not supported in Euphoria for %s"
		       {feature, version_name[LANG]}))
 end if                    
end procedure

global procedure InternalErr(object msg) -- kinz
-- Handles internal compile-time errors
-- see RTInternal() for run-time internal errors
    if TRANSLATE then
       if atom(msg) then
    screen_output(STDERR, sprintf(messages[022][LANG], --"Internal Error: %s\n"
				 {messages[msg][LANG]}))
       else
    screen_output(STDERR, sprintf(messages[022][LANG], --"Internal Error: %s\n"
						{msg}))
       end if
    else
       if atom(msg) then
 screen_output(STDERR, sprintf(messages[023][LANG], 
			      --"Internal Error at %s:%d - %s\n"
	   {file_name[current_file_no], line_number, messages[msg][LANG]}))
       else        
 screen_output(STDERR, sprintf(messages[023][LANG], 
			      --"Internal Error at %s:%d - %s\n"
	   {file_name[current_file_no], line_number, msg}))
       end if
    end if
    screen_output(STDERR, 018)--"\nPress Enter\n"
    if getc(0) then
    end if
    abort(1)
end procedure

