-- Public Domain software.
-- Euphoria 2.5 -- unlimited alphabet bilingual interpreter.
-- Any code page for the names.
-- English and Russian keywords, builtins, standard routines.
-- English  or Russian interpreter messages.
-- A Euphoria Interpreter written 100% in Euphoria,
-- reconstruction by Igor N. Kachan, 2005 May.
-- Report the problems to  kinz@peterlink.ru

-- Front End - main routine 

include file.e
include get.e
include pathoper.e

function ProcessOptions()
-- record command line options, return source file number
    integer src_file
    boolean dot_found
    sequence src_name, raw_name
    
    if Argc >= 3 then
	src_name = Argv[3]
    else 
	if INTERPRET and not BIND then
	    screen_output(STDERR, 001)--"Euphoria Interpreter " &
				      --INTERPRETER_VERSION & " "
	end if

	if ELINUX then
	    if BIND then
		screen_output(STDERR, 002)--"for Linux/FreeBSD.\n"
	    else    
		if EBSD then
		    screen_output(STDERR, 003)--"for FreeBSD.\n"
		else
		    screen_output(STDERR, 004)--"for Linux.\n"
		end if
	    end if
	    
	else
	    if BIND then
		screen_output(STDERR, 005)--"for DOS/Windows.\n"
	    elsif EWINDOWS then
		screen_output(STDERR, 006)--"for 32-bit Windows.\n"
	    else
		screen_output(STDERR, 007)--"for 32-bit DOS.\n"
	    end if
	end if
	
 machine_proc(6,#0607) -- underline_cursor
		
 screen_output
 (STDERR, 008)--"Source code provided by Rapid Deployment Software \n"
 screen_output
 (STDERR, 009)--"http://www.RapidEuphoria.com \n"
 screen_output
 (STDERR,
  010)--"Rapid Deployment Software is not responsible for any damage. \n")

	    screen_output(STDERR, 011)--"\nfile name to execute? "
	src_name = gets(STDIN)
	
	screen_output(STDERR, "\n")
    
	-- remove leading blanks
	while length(src_name) and find(src_name[1], " \t\n") do
	    src_name = src_name[2..$]
	end while
	
	if length(src_name) = 0 then
	    Cleanup(1)
	end if
	
	-- remove trailing blanks
	while length(src_name) and find(src_name[$], " \t\n") do
	    src_name = src_name[1..$-1]
	end while
	
	-- add src_name as 2nd arg for command_line()
	Argc = 2
	Argv = {Argv[1], src_name}
	
	file_name_entered = src_name -- passed to back-end for command_line()
	
	-- .ex or .exw might be added to src_name below
    end if
    
    -- check src_name for '.' */
    dot_found = FALSE
    for p = 1 to length(src_name) do
	if src_name[p] = '.' then
	   dot_found = TRUE
	elsif find(src_name[p], SLASH_CHARS) then
	   dot_found = FALSE
	end if
    end for 
    
    raw_name = src_name
    
    if not dot_found then
	-- no dot found */
	if ELINUX then
	    src_name &= ".exu"
	else
	    if (BIND and w32) or EWINDOWS then
		src_name &= ".exw"
	    else
		src_name &= ".ex"
	    end if
	end if
    end if

    file_name = append(file_name, src_name)
    src_file = e_path_open(src_name, "r")

    if ELINUX then
	if src_file = -1 then
	    -- do this especially so #! will work on Linux without .exu extension 
	    src_file = open(raw_name, "r")      
	    if src_file != -1 then
		src_name = raw_name
	    end if
	end if
    end if  
    return src_file
end function

function full_path(sequence filename)
-- return the full path for a file
    integer p
    p = length(filename)
    while p >= 1 do
	if find(filename[p], SLASH_CHARS) then
	    return filename[1..p]
	end if
	p -= 1
    end while
    return '.' & SLASH  
end function

procedure main()
-- Main routine 
    integer argc --,fn
    sequence argv
    
    -- we have our own, possibly different, idea of argc and argv
    argv = command_line()

    if BIND then
	argv = extract_options(argv)
    end if

    argc = length(argv)  

    Argv = argv
    Argc = argc
    
    eudir = getenv("EUDIR")
    if atom(eudir) then
	if ELINUX then
	    -- should check search PATH for euphoria/bin ?
	    eudir = getenv("HOME")
	    if atom(eudir) then
		eudir = "euphoria"  
	    else 
		eudir = eudir & "/euphoria"
	    end if
	else
	    eudir = "\\EUPHORIA"
	end if
    end if
    TempErrName = "ex.err"           

    
    src_file = ProcessOptions()

    if src_file = -1 then
	-- too early for normal error processing
screen_output(STDERR, sprintf(messages[091][LANG], {file_name[1]}))
			      --"Can't open %s\n"
	if BIND or EWINDOWS or ELINUX then
	    screen_output(STDERR, messages[092][LANG]) 
				  --"\nPress Enter\n"
	    if getc(0) then
	    end if
	end if
	Cleanup(1)
    end if
    
    main_path = full_path(file_name[1])  


    InitGlobals()

    InitSymTab()
    InitEmit()
    InitParser()
    InitLex()
    read_line()
    parser()

    -- we've parsed successfully
    -- now run the appropriate back-end 
    if TRANSLATE then
	BackEnd(0) -- translate IL to C
    
    elsif INTERPRET then----------------------
      --fn = open("code.ex", "wb")------------
	InitBackEnd(0)------------------------
	 --puts(fn, Code)---------------------
	 --close(fn)--------------------------
	
	BackEnd(0) -- execute IL using Euphoria-coded back-end
    end if

    Cleanup(0) -- does warnings

end procedure

main()

